// rhea-0.1.0
function build(type, attributes, ...children) {
    let element;
    if (typeof type == "function") {
        element = hydrate(type);
    }
    else {
        element = document.createElement(type);
    }
    if (attributes && typeof attributes == "string") {
        element.textContent = attributes;
    }
    else if (attributes && typeof attributes == "object" && attributes.text) {
        element.textContent = attributes.text;
    }
    if (typeof attributes == "object" && attributes != null) {
        Object.keys(attributes).forEach(item => {
            if (item == "text")
                return;
            if (element.hasAttribute(item) || item in element) {
                element.setAttribute(item, attributes[item]);
            }
            else if (item == "class") {
                element.classList.add(...attributes[item].split(" "));
            }
            else if (item.startsWith("data_")) {
                element.dataset[item.replace("data_", "")] = attributes[item];
            }
        });
    }
    let kids;
    if (children && typeof children[0] == "function") {
        kids = children.map((i) => {
            return i.call(null);
        });
    }
    else {
        kids = children;
    }
    if (children.length >= 1)
        append(element, ...kids);
    return element;
}
const NUMBER_OF_BATCHES = 25;
function append(parent, ...children) {
    const write = (kids) => {
        const frag = document.createDocumentFragment();
        kids.forEach((i) => {
            let el;
            if (i instanceof HTMLElement &&
                i instanceof HTMLAnchorElement &&
                i.href) {
                const url = new URL(i.href);
                if (url.origin.startsWith(window.location.origin) == true) {
                    i.addEventListener("click", evt => {
                        goTo(url, evt);
                    });
                }
            }
            else if (typeof i == "string") {
                el = document.createTextNode(i);
            }
            if (el != undefined) {
                frag.appendChild(el);
            }
            else {
                frag.appendChild(i);
            }
        });
        requestAnimationFrame(() => {
            if (!parent)
                parent = document.body;
            parent.appendChild(frag);
        });
    };
    if (children.length >= 250) {
        const distance = children.length / NUMBER_OF_BATCHES;
        let i = 0;
        while (i < children.length) {
            const r = i;
            setTimeout(() => write(children.slice(r, r + distance)), 0);
            i += distance;
        }
    }
    else {
        setTimeout(() => write(children), 0);
    }
    return parent;
}
function event(el, evtType, evt, opts, rd) {
    el.addEventListener(evtType, event => {
        evt.call(el, event);
    }, opts);
    if (el.dataset.component) {
        redraw(el.dataset.component);
        return;
    }
    if (rd) {
        redraw(rd);
    }
}
function head(...el) {
    el.forEach(i => {
        var _a;
        let selector = "";
        if (i.hasAttribute("title")) {
            selector = "title";
        }
        else if (i.hasAttribute("name")) {
            selector = "name";
        }
        else if (i.hasAttribute("href")) {
            selector = "href";
        }
        else if (i.hasAttribute("key")) {
            selector = "key";
        }
        let attr = "";
        if (selector.length >= 1) {
            attr = `[${selector}="${i.getAttribute(selector)}"]`;
        }
        (_a = document.head
            .querySelector(`${i.nodeName.toLowerCase()}${attr}`)) === null || _a === void 0 ? void 0 : _a.remove();
        document.head.appendChild(i);
    });
}
const batches = new Map();
let mutationQueued = false;
let latest = 0;
function mutate(element, props) {
    let el = element == null ? latest++ : element;
    batches.set(el, props);
    enqueue();
}
function enqueue() {
    if (mutationQueued == false) {
        mutationQueued = true;
        requestAnimationFrame(() => {
            for (const [el, fn] of batches) {
                if (el instanceof HTMLElement) {
                    const props = Object.entries(fn);
                    if (props.length >= 1) {
                        for (const [key, value] of props) {
                            if (typeof value == "object") {
                                for (const [prop, v] of Object.entries(value)) {
                                    el[key][prop] = v;
                                }
                            }
                            else {
                                el[key] = value;
                            }
                        }
                    }
                    else if (typeof fn == "function") {
                        fn.call(null, el);
                    }
                }
                else if (typeof el == "number" && typeof fn == "function") {
                    fn.call(null);
                }
                batches.delete(el);
                latest = 0;
            }
        });
        mutationQueued = false;
    }
}
const lazy = (src) => {
    return import(src).then(imports => {
        return imports.default;
    });
};

const requestAnimationFrame = window.requestAnimationFrame ||
    window.webkitRequestAnimationFrame ||
    window.mozRequestAnimationFrame ||
    window.msRequestAnimationFrame ||
    function (callback) {
        return setTimeout(callback, 1000 / 60);
    };
const Components = new Map();
const Index = new Map();
const NotFound = new Set();
window.addEventListener("popstate", () => {
    emit(document.body, EventType.BeforePageChange);
    render({}, true);
    emit(document.body, EventType.AfterPageChange);
});
const register = (...elements) => {
    return elements
        .map(element => {
        const n = element.name.toLowerCase();
        if (Components.get(n)) {
            throw Error(`component ${n} has already been defined!`);
        }
        else {
            const ob = {
                fn: element,
                el: element.call(null),
            };
            Components.set(element.name.toLowerCase(), ob);
            return true;
        }
    })
        .some(i => (i == false ? false : true));
};
const mount = (route, page) => {
    const fn = page.call(null).name.toLowerCase();
    if (route == undefined) {
        NotFound.add(fn);
    }
    else {
        Index.set(route, new Set().add(fn));
    }
    return true;
};
const removeAll = (el) => {
    requestAnimationFrame(() => {
        Array.from(el).forEach(i => i.remove());
    });
};
const navigate = (destination) => {
    const old = window.location.pathname;
    const p = typeof destination == "string" ? destination : destination.pathname;
    if (p != old)
        window.history.pushState("", "", p);
    emit(document.body, EventType.Navigation, {
        old,
        next: p,
    });
    render({}, true);
    emit(document.body, EventType.AfterNavigationChange);
};
var EventType;
(function (EventType) {
    EventType["Initialized"] = "initialized";
    EventType["Navigation"] = "navigation";
    EventType["Render"] = "render";
    EventType["Redraw"] = "redraw";
    EventType["GlobalRender"] = "global-render";
    EventType["BeforePageChange"] = "before-page-change";
    EventType["AfterPageChange"] = "after-page-change";
    EventType["AfterNavigationChange"] = "after-navigation-change";
})(EventType || (EventType = {}));
const emit = (el, event, detail) => el.dispatchEvent(new CustomEvent(event.toString(), { detail }));
const goTo = (url, evt) => {
    evt === null || evt === void 0 ? void 0 : evt.preventDefault();
    navigate(url);
};
const path = (destination = window.location.pathname) => {
    var _a;
    const r = (_a = Array.from(Index)) === null || _a === void 0 ? void 0 : _a.find(([i, s]) => {
        if (typeof i == "string" && i == destination) {
            return s;
        }
        else if (i instanceof RegExp && i.test(destination) == true) {
            return s;
        }
        else {
            return false;
        }
    });
    if (r && r[0])
        return r[1];
    else {
        return false;
    }
};
const prerender = (route, set = false) => {
    let frag = new DocumentFragment();
    let components;
    if (route == undefined) {
        components = NotFound;
    }
    else {
        components = Index.get(route);
    }
    const h = () => {
        if (components == undefined)
            return;
        components.forEach(i => {
            const cmp = Components.get(i);
            if (cmp) {
                frag.appendChild(hydrate(cmp.fn));
            }
            else {
                throw Error(`component ${i} is not registered in the component registry`);
            }
        });
    };
    if (components != undefined) {
        h();
    }
    else {
        if (route instanceof RegExp) {
            Index.forEach((_, v) => {
                if (v instanceof RegExp) {
                    if (v.source === route.source &&
                        v.global === route.global &&
                        v.ignoreCase === route.ignoreCase &&
                        v.multiline === route.multiline) {
                        components = Index.get(v);
                        h();
                    }
                }
            });
        }
        else if (typeof route == "string") {
            Index.forEach((_, v) => {
                if (v instanceof RegExp && v.test(route) == true) {
                    components = Index.get(v);
                    h();
                }
            });
        }
    }
    if (set == true && cache.has(route) == false)
        cache.set(route, frag);
    return frag;
};
const cache = new Map();
const render = (options, prev = true, route = window.location.pathname) => {
    if (prev)
        removeAll(document.body.children);
    const r = (components, route) => {
        let frag = new DocumentFragment();
        const cached = cache.get(route);
        if (cached == undefined) {
            components.forEach(i => {
                const cmp = Components.get(i);
                if (cmp) {
                    frag.appendChild(hydrate(cmp.fn));
                }
                else {
                    throw Error(`component ${i} is not registered in the component registry`);
                }
            });
        }
        else {
            frag = cached;
        }
        requestAnimationFrame(() => {
            document.body.append(frag);
            emit(document.body, EventType.GlobalRender);
        });
    };
    if (options === null || options === void 0 ? void 0 : options.prerender)
        options.prerender.forEach(i => window.requestIdleCallback(() => prerender(i, true), {
            timeout: 100,
        }));
    const components = path(route);
    if (components == false && NotFound.size == 0) {
        throw Error(`${window.location.pathname} is not a registered route`);
    }
    else if (NotFound.size != 0 && components == false) {
        r(NotFound, undefined);
        return true;
    }
    else if (components != false) {
        r(components);
        return true;
    }
    else {
        throw Error(`${window.location.pathname} is an invalid route`);
    }
};
const hydrate = (i) => {
    const el = i === null || i === void 0 ? void 0 : i.call(null);
    emit(document.body, EventType.Initialized, {
        component: i.name.toLowerCase(),
    });
    const kids = Array.from(el.children).map(el => el);
    const hasLink = kids.filter((kid) => kid instanceof HTMLAnchorElement &&
        kid.href &&
        new URL(kid.href).origin.startsWith(window.location.origin) == true);
    if (hasLink.length >= 1) {
        hasLink.forEach(i => i.addEventListener("click", evt => {
            if (i.dataset.bound)
                return;
            goTo(new URL(i.href), evt);
        }));
    }
    if (el instanceof HTMLAnchorElement &&
        new URL(el.href).origin.startsWith(window.location.origin) == true) {
        el.addEventListener("click", evt => {
            if (el.dataset.bound)
                return;
            el.dataset.bound = "true";
            goTo(new URL(el.href), evt);
        });
    }
    el.dataset.component = i.name.toLowerCase();
    return el;
};
const redraw = (cmp) => {
    let component = cmp;
    if (typeof cmp == "function") {
        component = cmp.name.toLowerCase();
    }
    if (typeof component == "function") {
        component = component.name.toLowerCase();
    }
    const cmpFunction = Components.get(component);
    if (cmpFunction) {
        const { el, fn } = cmpFunction;
        el.replaceWith(hydrate(fn));
        emit(el, EventType.Redraw);
    }
    else {
        throw Error(`component ${component} is not registered`);
    }
};

export { Components, EventType, Index, NotFound, append, build, cache, event, goTo, head, hydrate, lazy, mount, mutate, navigate, prerender, redraw, register, render, requestAnimationFrame };
//# sourceMappingURL=rhea.es2017.js.map
